(ns puppetlabs.rbac.services.authn
  "This namespace exposes functions that are primarily of interest
  only to the rbac-ui (though it may be expanded post-3.4). It
  provides the login! and logout! functions needed to implement
  authentication pages, handling every aspect of authentication except
  the core Shiro calls (which are handled in
  puppetlabs.rbac.shiro-utils).

  Development versions of authentication pages are housed in
  puppetlabs.rbac.testutils.services.dev-login and should be used as
  an implementation reference."
  (:require [ring.util.codec :refer [url-decode]]
            [slingshot.slingshot :refer [throw+]]
            [puppetlabs.rbac.shiro-utils :as shiro]
            [puppetlabs.rbac.storage :as st]
            [puppetlabs.rbac.sync :as sn]
            [puppetlabs.trapperkeeper.core :refer [defservice]]
            [puppetlabs.trapperkeeper.services :refer [service-context]]))

(defn strip-protocol
  "Removes leading protocol:// from a url"
  [url]
  (clojure.string/replace url #".*://" ""))

(defn relativize
  "Converts url into relative url."
  [url]
  (if (re-matches #"^/.*$" url)
    url
    (->> (clojure.string/split url #"/")
         rest
         (clojure.string/join "/")
         (str "/"))))

(defn do-sanitize-redirect
  "Given a redirect URL, ensure that it is safe for a redirect (and
  not a phishing attempt) by making it relative."
  [url]
  (-> url
      url-decode
      strip-protocol
      relativize))

(defn rbac-login!
  "Authenticates a user using Shiro. Takes a username, password, shiro-subject
  (the value of the shiro-subject key embedded into the request by the middleware),
  and the database spec."
  [username password shiro-subject db-spec max-failed-logins]
  (let [rbac-subject (st/get-subject db-spec "login" username)]

    (when (:is_revoked rbac-subject)
      (throw+ {:kind :puppetlabs.rbac/user-revoked
               :msg "User is revoked."}))

    ;; TODO there may be a more elegant thing to do, here. It
    ;; might also become more evident what to do once we have
    ;; the unauthenticated assets server in place.
    (when (shiro/authenticated? shiro-subject)
      (shiro/logout! shiro-subject))

    (try (shiro/login! shiro-subject username password)
         ;; TODO need to preserve redirect, see PE-5686
         (catch org.apache.shiro.authc.AuthenticationException e
           (st/update-failed-login db-spec username max-failed-logins)
           (throw+ {:kind :puppetlabs.rbac/user-unauthenticated
                    :msg "Authentication failed."})))

    (if rbac-subject
      (when (:is_remote rbac-subject)
        (sn/sync-remote-user! db-spec (:id rbac-subject)))
      (sn/create-remote-user! db-spec username))

    (st/update-last-login db-spec username)

    nil))

(defn rbac-logout! [shiro-subject]
  (shiro/logout! shiro-subject)
  nil)

(defprotocol RbacAuthnService
  "Protocol for authentication with RBAC."
  (sanitize-redirect [this url])
  (login!  [this username password shiro-subject])
  (logout! [this shiro-subject]))

(defservice rbac-authn-service
  "A service which can be used to log a user in or out using RBAC."
  RbacAuthnService
  [[:ConfigService get-in-config]]
  (init [this context]
        (let [db (get-in-config [:rbac-database])
              global-config (get-in-config [:global])]
          (assoc context
            :db db
            :failed-attempts-lockout (:failed-attempts-lockout global-config 10))))

  (sanitize-redirect [this url] (do-sanitize-redirect url))

  (login! [this username password shiro-subject]
          (rbac-login! username password shiro-subject
                       (:db (service-context this))
                       (:failed-attempts-lockout (service-context this))))

  (logout! [this shiro-subject]
           (rbac-logout! shiro-subject)))
